import enum
import platform
import os

import copy
from aardwolf.commons.credential import *
from aardwolf.commons.proxy import  RDPProxyType
from aardwolf.authentication.spnego.native import SPNEGO
from aardwolf.authentication.ntlm.native import NTLMAUTHHandler, NTLMHandlerSettings
from aardwolf.authentication.kerberos.native import RDPKerberos
from aardwolf.authentication.credssp.native import CredSSPAuth
from minikerberos.common.target import KerberosTarget
from minikerberos.common.proxy import KerberosProxy
from minikerberos.common.creds import KerberosCredential
from minikerberos.common.spn import KerberosSPN

from aardwolf.commons.target import RDPTarget


if platform.system().upper() == 'WINDOWS':
	from aardwolf.authentication.kerberos.sspi import RDPKerberosSSPI
	from aardwolf.authentication.ntlm.sspi import RDPNTLMSSPI

class AuthenticatorBuilder:
	def __init__(self):
		pass
	
	@staticmethod
	def wrap_credssp(spnego, credtype, cred):
		if credtype.find('KERBEROS') != -1:
			ntlm = spnego
		else:
			ntlm = spnego.get_original_context('NTLMSSP - Microsoft NTLM Security Support Provider')
		credssp_ctx = CredSSPAuth()
		credssp_ctx.auth_ctx = ntlm
		credssp_ctx.credtype = credtype
		credssp_ctx.cred = cred
		return credssp_ctx



	@staticmethod
	def to_credssp(creds:RDPCredential, target:RDPTarget = None):
		if creds.authentication_type == RDPAuthProtocol.PLAIN:
			ntlmcred = RDPNTLMCredential()
			ntlmcred.username = creds.username
			ntlmcred.domain = creds.domain if creds.domain is not None else ''
			ntlmcred.workstation = None
			ntlmcred.is_guest = False
			
			if creds.secret is None:
				if creds.username is None and creds.domain is None:
					ntlmcred.is_guest = True
				else:
					raise Exception('Authentication requres password!')
			
			ntlmcred.password = creds.secret
			
			settings = NTLMHandlerSettings(ntlmcred)
			handler = NTLMAUTHHandler(settings)
			
			#setting up SPNEGO
			spneg = SPNEGO()
			spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
			
			return AuthenticatorBuilder.wrap_credssp(spneg, 'NTLM', ntlmcred)

		elif creds.authentication_type == RDPAuthProtocol.NTLM:
			ntlmcred = RDPNTLMCredential()
			ntlmcred.username = creds.username
			ntlmcred.domain = creds.domain if creds.domain is not None else ''
			ntlmcred.workstation = None
			ntlmcred.is_guest = False
			
			if creds.secret is None:
				if creds.username is None and creds.domain is None:
					ntlmcred.is_guest = True
				else:
					raise Exception('NTLM authentication requres password!')
				
			if creds.secret_type == RDPCredentialsSecretType.NT:
				if isinstance(creds.secret, str) is True and len(creds.secret) != 32:
					raise Exception('This is not an NT hash!')
				ntlmcred.nt_hash = creds.secret
			elif creds.secret_type == RDPCredentialsSecretType.PASSWORD:
				ntlmcred.password = creds.secret
			
			settings = NTLMHandlerSettings(ntlmcred)
			handler = NTLMAUTHHandler(settings)
			
			#setting up SPNEGO
			spneg = SPNEGO()
			spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
			
			return AuthenticatorBuilder.wrap_credssp(spneg, 'NTLM', ntlmcred)
		
		elif creds.authentication_type == RDPAuthProtocol.KERBEROS:
			if target is None:
				raise Exception('Target must be specified with Kerberos!')
				
			if target.hostname is None:
				raise Exception('target must have a domain name or hostname for kerberos!')
				
			if target.dc_ip is None:
				raise Exception('target must have a dc_ip for kerberos!')
			
			if creds.secret_type == RDPCredentialsSecretType.KEYTAB:
				filename = creds.secret
				if creds.secret.upper() == 'ENV':
					filename = os.environ['KRB5KEYTAB']

				kc = KerberosCredential.from_keytab(filename, creds.username, creds.domain)

			elif creds.secret_type == RDPCredentialsSecretType.CCACHE:
				filename = creds.secret
				if creds.secret.upper() == 'ENV':
					try:
						filename = os.environ['KRB5CCACHE']
					except:
						raise Exception('Kerberos auth missing environment variable KRB5CCACHE')
				kc = KerberosCredential.from_ccache_file(filename)
				kc.username = creds.username
				kc.domain = creds.domain
			
			elif creds.secret_type == RDPCredentialsSecretType.KIRBI:
				filename = creds.secret
				kc = KerberosCredential.from_kirbi(filename)
				kc.username = creds.username
				kc.domain = creds.domain
			
			else:
				kc = KerberosCredential()
				kc.username = creds.username
				kc.domain = creds.domain
				if creds.secret_type == RDPCredentialsSecretType.PASSWORD:
					kc.password = creds.secret
				elif creds.secret_type == RDPCredentialsSecretType.NT:
					kc.nt_hash = creds.secret
					
				elif creds.secret_type == RDPCredentialsSecretType.AES:
					if len(creds.secret) == 32:
						kc.kerberos_key_aes_128 = creds.secret
					elif len(creds.secret) == 64:
						kc.kerberos_key_aes_256 = creds.secret
						
				elif creds.secret_type == RDPCredentialsSecretType.RC4:
					kc.kerberos_key_rc4 = creds.secret

			if kc is None:
				raise Exception('No suitable secret type found to set up kerberos!')
			
				
			kcred = RDPKerberosCredential()
			kcred.ccred = kc
			kcred.spn = KerberosSPN.from_target_string(target.to_target_string())
			
			if target.proxy is not None:
				if target.proxy.type == RDPProxyType.ASYSOCKS:
					kcred.target = KerberosTarget(target.dc_ip)
					kcred.target.proxy = KerberosProxy()
					kcred.target.proxy.target = copy.deepcopy(target.proxy.target)
					kcred.target.proxy.target[-1].endpoint_ip = target.dc_ip
					kcred.target.proxy.target[-1].endpoint_port = 88
				
				else:
					raise NotImplementedError("Proxy type '%s' not implemented!" % target.proxy.type)

			else:
				kcred.target = KerberosTarget(target.dc_ip)
			handler = RDPKerberos(kcred)
			#setting up SPNEGO
			spneg = SPNEGO()
			spneg.add_auth_context('MS KRB5 - Microsoft Kerberos 5', handler)
			return AuthenticatorBuilder.wrap_credssp(spneg, 'KERBEROS', kc)
			
		elif creds.authentication_type == RDPAuthProtocol.SSPI_KERBEROS:
			if target is None:
				raise Exception('Target must be specified with Kerberos SSPI!')
				
			kerbcred = RDPKerberosSSPICredential()
			kerbcred.client = None #creds.username #here we could submit the domain as well for impersonation? TODO!
			kerbcred.password = creds.secret
			kerbcred.target = target.to_target_string()
			
			handler = RDPKerberosSSPI(kerbcred)
			#setting up SPNEGO
			spneg = SPNEGO()
			spneg.add_auth_context('MS KRB5 - Microsoft Kerberos 5', handler)
			return AuthenticatorBuilder.wrap_credssp(spneg, 'SSPI-KERBEROS', kerbcred)
		
		elif creds.authentication_type == RDPAuthProtocol.SSPI_NTLM:
			ntlmcred = RDPNTLMSSPICredential()
			ntlmcred.client = creds.username #here we could submit the domain as well for impersonation? TODO!
			ntlmcred.password = creds.secret
			
			handler = RDPNTLMSSPI(ntlmcred)
			#setting up SPNEGO
			spneg = SPNEGO()
			spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
			return AuthenticatorBuilder.wrap_credssp(spneg, 'SSPI-NTLM', ntlmcred)

		elif creds.authentication_type.value.startswith('MPN'):
			if creds.authentication_type in [RDPAuthProtocol.MPN_SSL_NTLM, RDPAuthProtocol.MPN_NTLM]:
				from aardwolf.authentication.ntlm.mpn import RDPNTLMMPN
				ntlmcred = RDPMPNCredential()
				ntlmcred.type = 'NTLM'
				ntlmcred.is_ssl = True if creds.authentication_type == RDPAuthProtocol.MPN_SSL_NTLM else False
				ntlmcred.parse_settings(creds.settings)
				
				handler = RDPNTLMMPN(ntlmcred)
				#setting up SPNEGO
				spneg = SPNEGO()
				spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
				return AuthenticatorBuilder.wrap_credssp(spneg, 'MPN-NTLM', ntlmcred)

			elif creds.authentication_type in [RDPAuthProtocol.MPN_SSL_KERBEROS, RDPAuthProtocol.MPN_KERBEROS]:
				from aardwolf.authentication.kerberos.mpn import RDPKerberosMPN

				ntlmcred = RDPMPNCredential()
				ntlmcred.type = 'KERBEROS'
				ntlmcred.target = creds.target
				if creds.username is not None:
					ntlmcred.username = '<CURRENT>'
				if creds.domain is not None:
					ntlmcred.domain = '<CURRENT>'
				if creds.secret is not None:
					ntlmcred.password = '<CURRENT>'
				ntlmcred.is_guest = False
				ntlmcred.is_ssl = True if creds.authentication_type == RDPAuthProtocol.MPN_SSL_KERBEROS else False
				ntlmcred.parse_settings(creds.settings)

				handler = RDPKerberosMPN(ntlmcred)
				#setting up SPNEGO
				spneg = SPNEGO()
				spneg.add_auth_context('MS KRB5 - Microsoft Kerberos 5', handler)
				return AuthenticatorBuilder.wrap_credssp(spneg, 'MPN-KERBEROS', ntlmcred)

				

		elif creds.authentication_type.value.startswith('MULTIPLEXOR'):
			if creds.authentication_type in [RDPAuthProtocol.MULTIPLEXOR_SSL_NTLM, RDPAuthProtocol.MULTIPLEXOR_NTLM]:
				from aardwolf.authentication.ntlm.multiplexor import RDPNTLMMultiplexor

				ntlmcred = RDPMultiplexorCredential()
				ntlmcred.type = 'NTLM'
				if creds.username is not None:
					ntlmcred.username = '<CURRENT>'
				if creds.domain is not None:
					ntlmcred.domain = '<CURRENT>'
				if creds.secret is not None:
					ntlmcred.password = '<CURRENT>'
				ntlmcred.is_guest = False
				ntlmcred.is_ssl = True if creds.authentication_type == RDPAuthProtocol.MULTIPLEXOR_SSL_NTLM else False
				ntlmcred.parse_settings(creds.settings)
				
				handler = RDPNTLMMultiplexor(ntlmcred)
				#setting up SPNEGO
				spneg = SPNEGO()
				spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
				return AuthenticatorBuilder.wrap_credssp(spneg, 'MULTIPLEXOR-NTLM', ntlmcred)

			elif creds.authentication_type in [RDPAuthProtocol.MULTIPLEXOR_SSL_KERBEROS, RDPAuthProtocol.MULTIPLEXOR_KERBEROS]:
				from aardwolf.authentication.kerberos.multiplexor import RDPKerberosMultiplexor

				ntlmcred = RDPMultiplexorCredential()
				ntlmcred.type = 'KERBEROS'
				ntlmcred.target = creds.target
				if creds.username is not None:
					ntlmcred.username = '<CURRENT>'
				if creds.domain is not None:
					ntlmcred.domain = '<CURRENT>'
				if creds.secret is not None:
					ntlmcred.password = '<CURRENT>'
				ntlmcred.is_guest = False
				ntlmcred.is_ssl = True if creds.authentication_type == RDPAuthProtocol.MULTIPLEXOR_SSL_NTLM else False
				ntlmcred.parse_settings(creds.settings)

				handler = RDPKerberosMultiplexor(ntlmcred)
				#setting up SPNEGO
				spneg = SPNEGO()
				spneg.add_auth_context('MS KRB5 - Microsoft Kerberos 5', handler)
				return AuthenticatorBuilder.wrap_credssp(spneg, 'MULTIPLEXOR-KERBEROS', ntlmcred)

		elif creds.authentication_type.value.startswith('WSNET'):
			if creds.authentication_type in [RDPAuthProtocol.WSNET_NTLM]:
				from aardwolf.authentication.ntlm.wsnet import RDPWSNetNTLMAuth
				
				ntlmcred = RDPWSNETCredential()
				ntlmcred.type = 'NTLM'
				if creds.username is not None:
					ntlmcred.username = '<CURRENT>'
				if creds.domain is not None:
					ntlmcred.domain = '<CURRENT>'
				if creds.secret is not None:
					ntlmcred.password = '<CURRENT>'
				ntlmcred.is_guest = False
				
				handler = RDPWSNetNTLMAuth(ntlmcred)
				spneg = SPNEGO()
				spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
				return AuthenticatorBuilder.wrap_credssp(spneg, 'WSNET-NTLM', ntlmcred)
			

			elif creds.authentication_type in [RDPAuthProtocol.WSNET_KERBEROS]:
				from aardwolf.authentication.kerberos.wsnet import RDPWSNetKerberosAuth

				ntlmcred = RDPWSNETCredential()
				ntlmcred.type = 'KERBEROS'
				ntlmcred.target = creds.target
				if creds.username is not None:
					ntlmcred.username = '<CURRENT>'
				if creds.domain is not None:
					ntlmcred.domain = '<CURRENT>'
				if creds.secret is not None:
					ntlmcred.password = '<CURRENT>'
				ntlmcred.is_guest = False

				handler = RDPWSNetKerberosAuth(ntlmcred)
				#setting up SPNEGO
				spneg = SPNEGO()
				spneg.add_auth_context('MS KRB5 - Microsoft Kerberos 5', handler)
				return AuthenticatorBuilder.wrap_credssp(spneg, 'WSNET-KERBEROS', ntlmcred)
		
		elif creds.authentication_type.value.startswith('SSPIPROXY'):
			if creds.authentication_type == RDPAuthProtocol.SSPIPROXY_NTLM:
				from aardwolf.authentication.ntlm.sspiproxy import RDPSSPIProxyNTLMAuth
				
				ntlmcred = RDPSSPIProxyCredential()
				ntlmcred.type = 'NTLM'
				if creds.username is not None:
					ntlmcred.username = '<CURRENT>'
				if creds.domain is not None:
					ntlmcred.domain = '<CURRENT>'
				if creds.secret is not None:
					ntlmcred.password = '<CURRENT>'
				ntlmcred.is_guest = False
				ntlmcred.host = creds.settings['host'][0]
				ntlmcred.port = int(creds.settings['port'][0])
				ntlmcred.proto = 'ws'
				if 'proto' in creds.settings:
					ntlmcred.proto = creds.settings['proto'][0]
				if 'agentid' in creds.settings:
					ntlmcred.agent_id = bytes.fromhex(creds.settings['agentid'][0])
				
				handler = RDPSSPIProxyNTLMAuth(ntlmcred)
				spneg = SPNEGO()
				spneg.add_auth_context('NTLMSSP - Microsoft NTLM Security Support Provider', handler)
				return AuthenticatorBuilder.wrap_credssp(spneg, 'SSPIPROXY-NTLM', ntlmcred)
			

			elif creds.authentication_type == RDPAuthProtocol.SSPIPROXY_KERBEROS:
				from aardwolf.authentication.kerberos.sspiproxy import RDPSSPIProxyKerberosAuth

				ntlmcred = RDPSSPIProxyCredential()
				ntlmcred.type = 'KERBEROS'
				ntlmcred.target = creds.target
				if creds.username is not None:
					ntlmcred.username = '<CURRENT>'
				if creds.domain is not None:
					ntlmcred.domain = '<CURRENT>'
				if creds.secret is not None:
					ntlmcred.password = '<CURRENT>'
				ntlmcred.is_guest = False
				ntlmcred.host = creds.settings['host'][0]
				ntlmcred.port = int(creds.settings['port'][0])
				ntlmcred.proto = 'ws'
				if 'proto' in creds.settings:
					ntlmcred.proto = creds.settings['proto'][0]
				if 'agentid' in creds.settings:
					ntlmcred.agent_id = bytes.fromhex(creds.settings['agentid'][0])

				handler = RDPSSPIProxyKerberosAuth(ntlmcred)
				#setting up SPNEGO
				spneg = SPNEGO()
				spneg.add_auth_context('MS KRB5 - Microsoft Kerberos 5', handler)
				return AuthenticatorBuilder.wrap_credssp(spneg, 'SSPIPROXY-KERBEROS', ntlmcred)
		



import os
from aardwolf import logger
from aardwolf.authentication.spnego.native import SPNEGO
from aardwolf.authentication.credssp.messages.asn1_structs import *
from hashlib import sha256

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cssp/385a7489-d46b-464c-b224-f7340e308a5c

class CredSSPAuth:
	def __init__(self, mode = 'CLIENT'):
		self.mode = mode
		self.auth_ctx:SPNEGO = None
		self.credtype:str = None
		self.cred = None
		self.version = 4 #ver 5 and 6 not working TODO
		self.nonce = os.urandom(32)
		self.__internal_auth_continue = True
		self.seqno = 0

	def get_extra_info(self):
		return self.auth_ctx.get_extra_info()

	async def authenticate(self, token, flags = None, pubkey = None, remote_credguard = False):
		try:
			# currently only SSPI supported

			if self.mode != 'CLIENT':
				raise NotImplementedError()
			else:
				if token is None:
					# initial auth
					returndata, self.__internal_auth_continue, err = await self.auth_ctx.authenticate(token, flags = flags, seq_number = None)
					if err is not None:
						raise err
					
					negotoken = {
						'negoToken' : returndata
					}
					retoken = {
						'version' : self.version,
						'negoTokens' : NegoDatas([NegoData(negotoken)])
					}
					return TSRequest(retoken).dump(), True, None
				else:
					if self.__internal_auth_continue is True:
						tdata = TSRequest.load(token)
						if tdata.native['version'] < self.version:
							logger.debug('[CREDSSP] Server supports version %s which is smaller than our supported version %s' % (tdata.native['version'], self.version))
							self.version = tdata.native['version']
						if tdata.native['negoTokens'] is None:
							raise Exception('SSPI auth not supported by server')
						sspitoken = tdata.native['negoTokens'][0]['negoToken']
						returndata, self.__internal_auth_continue, err = await self.auth_ctx.authenticate(sspitoken, flags = flags)
						if err is not None:
							raise err
						
						negotoken = {
							'negoToken' : returndata
						}
						retoken = {
							'version' : self.version,
						}
						if returndata is not None:
							retoken['negoTokens'] = NegoDatas([NegoData(negotoken)])
						
						
						if self.__internal_auth_continue is False:
							if self.version in [5,6]:
								ClientServerHashMagic = b"CredSSP Client-To-Server Binding Hash\x00"
								ClientServerHash = sha256(ClientServerHashMagic + self.nonce + pubkey).digest()
								sealedMessage, signature = await self.auth_ctx.encrypt(ClientServerHash, self.seqno)
								self.seqno += 1
								#print(sealedMessage)
								retoken['pubKeyAuth'] = signature+sealedMessage
								retoken['clientNonce'] = self.nonce
							
							elif self.version in [2,3,4]:
								sealedMessage, signature = await self.auth_ctx.encrypt(pubkey, self.seqno)
								self.seqno += 1
								#print(sealedMessage)
								retoken['pubKeyAuth'] = signature+sealedMessage
						
						return TSRequest(retoken).dump(), True, None
					else:
						# sub-level auth protocol finished, now for the other stuff
						
						# waiting for server to reply with the re-encrypted verification string + b'\x01'
						tdata = TSRequest.load(token).native
						if tdata['errorCode'] is not None:
							raise Exception('CredSSP - Server sent an error! Code: %s' % tdata['errorCode'])
						if tdata['pubKeyAuth'] is None:
							raise Exception('Missing pubKeyAuth')
						verification_data, _ = await self.auth_ctx.decrypt(tdata['pubKeyAuth'], 0)
						#print('DEC: %s' % verification_data)

						# at this point the verification should be implemented
						# TODO: maybe later...

						
						# sending credentials
						if remote_credguard is False:
							creds = {
								'domainName' : b'',
								'userName'   : b'',
								'password'   : b'',
							}
							if self.cred.password is not None:
								creds = {
									'domainName' : self.cred.domain.encode('utf-16-le'),
									'userName'   : self.cred.username.encode('utf-16-le'),
									'password'   : self.cred.password.encode('utf-16-le'),
								}							 
								 
										
							res = TSPasswordCreds(creds)
							res = TSCredentials({'credType': 1, 'credentials': res.dump()})
							sealedMessage, signature = await self.auth_ctx.encrypt(res.dump(), self.seqno) #seq number must be incremented here..
							self.seqno += 1
							retoken = {
								'version' : self.version,
								'authInfo' : signature+sealedMessage
							}
							return TSRequest(retoken).dump(), False, None
						else:
							credBuffer = b''
							data = {
								'packageName' : 'KERBEROS'.encode('utf-16-le'), #dont forget the encoding!
								'credBuffer' : credBuffer
							}

							remcredguardcreds = TSRemoteGuardCreds({
								'logonCred' : TSRemoteGuardPackageCred(data),
								#'supplementalCreds' : [TSRemoteGuardPackageCred(xxxx)]
							})

							#print(remcredguardcreds)
							sealedMessage, signature = await self.auth_ctx.encrypt(remcredguardcreds.dump(), self.seqno) #seq number must be incremented here..
							self.seqno += 1

							retoken = {
								'version' : self.version,
								'authInfo' : signature+sealedMessage
							}
							return TSRequest(retoken).dump(), False, None
						
		except Exception as e:
			return None, None, e


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
# Comments:
#


import ipaddress
import enum
import copy
from aardwolf.commons.proxy import RDPProxy

class RDPConnectionDialect(enum.Enum):
	RDP = 'RDP'
	VNC = 'VNC'

class RDPConnectionProtocol(enum.Enum):
	TCP = 'TCP'

class RDPTarget:
	"""
	"""
	def __init__(
			self, 
			ip:str = None, 
			port:int = 3389, 
			hostname:str = None, 
			timeout:int = 1, 
			dc_ip:int=None, 
			domain:str = None, 
			proxy:RDPProxy = None, 
			protocol:RDPConnectionProtocol = RDPConnectionProtocol.TCP, 
			serverip = None, 
			dialect:RDPConnectionDialect = RDPConnectionDialect.RDP):
		
		self.ip = ip
		self.port = port
		self.hostname = hostname
		self.timeout = timeout
		self.dc_ip = dc_ip
		self.domain = domain
		self.proxy = proxy
		self.protocol = protocol
		self.serverip = serverip
		self.dialect = dialect
		if self.dialect == RDPConnectionDialect.VNC:
			self.port = 5900

	def to_target_string(self):
		return 'termsrv/%s@%s' % (self.hostname, self.domain)

	def get_copy(self, ip, port, hostname = None):
		t = RDPTarget(
			ip = ip, 
			port = port, 
			hostname = hostname, 
			timeout = self.timeout, 
			dc_ip= self.dc_ip, 
			domain = self.domain, 
			proxy = copy.deepcopy(self.proxy),
			protocol = self.protocol,
			serverip = self.serverip,
			dialect = self.dialect
		)
		return t
	
	@staticmethod
	def from_connection_string(s, is_rdp = True):
		port = 3389
		if is_rdp is False:
			port = 5900
		
		dc = None
		
		_, target = s.rsplit('@', 1)
		if target.find('/') != -1:
			target, dc = target.split('/')
			
		if target.find(':') != -1:
			target, port = target.split(':')
			
		st = RDPTarget()
		st.port = port
		st.dc_ip = dc
		st.domain, _ = s.split('/', 1)
		
		try:
			st.ip = str(ipaddress.ip_address(target))
		except:
			st.hostname = target
	
		return st
		
	def get_ip(self):
		if not self.ip and not self.hostname:
			raise Exception('RDPTarget must have ip or hostname defined!')
		return self.ip if self.ip is not None else self.hostname
		
	def get_hostname(self):
		return self.hostname
	
	def get_hostname_or_ip(self):
		if self.hostname:
			return self.hostname
		return self.ip
	
	def get_port(self):
		return self.port
		
	def __str__(self):
		t = '==== RDPTarget ====\r\n'
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
			
		return t
		
		
def test():
	s = 'TEST/victim/ntlm/nt:AAAAAAAA@10.10.10.2:3389'
	creds = RDPTarget.from_connection_string(s)
	print(str(creds))
	
	s = 'TEST/victim/sspi@10.10.10.2:3389/aaaa'
	creds = RDPTarget.from_connection_string(s)
	
	print(str(creds))
	
if __name__ == '__main__':
	test()


import asyncio

import platform
import traceback
try:
	import ssl
except:
	if platform.system() == 'Emscripten':
		#pyodide doesnt support SSL for now.
		pass


import logging
from asn1crypto.x509 import Certificate, PublicKeyInfo

class SSLClientTunnel:
	def __init__(self):
		self.in_queue = None
		self.out_queue = None
		self.ssl_ctx = None
		self.transport = None
		self.ssl_in_buff = None
		self.ssl_out_buff = None
		self.ssl_obj = None
		self.__ssl_in_q = None
		self.stop_evt = None
		self.server_certificate = None

		self.__enc_task = None
		self.__dec_task = None
		self.__process_task = None
	
	async def disconnect(self):
		try:
			if self.__enc_task is not None:
				self.__enc_task.cancel()
			if self.__dec_task is not None:
				self.__dec_task.cancel()
			if self.__process_task is not None:
				self.__process_task.cancel()

			if self.transport is not None:
				self.transport.out_queue.put_nowait(b'')
				await self.transport.disconnect()
		except Exception as e:
			pass
	
	async def get_server_pubkey(self):
		try:
			if self.server_certificate is None:
				return None
			cert = Certificate.load(self.server_certificate)
			return cert['tbs_certificate']['subject_public_key_info']['public_key'].dump()[5:] #why?
		except Exception as e:
			return e
	
	async def setup(self):
		try:
			self.in_queue = asyncio.Queue()
			self.out_queue = asyncio.Queue()
			self.stop_evt = asyncio.Event()
			self.__ssl_in_q = asyncio.Queue()
			self.ssl_in_buff = ssl.MemoryBIO()
			self.ssl_out_buff = ssl.MemoryBIO()
			self.ssl_obj = self.ssl_ctx.wrap_bio(self.ssl_in_buff, self.ssl_out_buff, server_side=False, server_hostname = None)
			_, err = await self.__do_handshake()
			if err is not None:
				raise err
			#print('Handshake ok!')
			self.server_certificate = self.ssl_obj.getpeercert(binary_form=True)
			self.__enc_task = asyncio.create_task(self.__encrypt_ssl_rec())
			self.__dec_task = asyncio.create_task(self.__decrypt_ssl_rec())
			self.__process_task = asyncio.create_task(self.__read_ssl_record())
			return True, None

		except Exception as e:
			return None, e
	
	async def __encrypt_ssl_rec(self):
		try:
			while True:
				try:
					raw_data = await self.out_queue.get()
					if raw_data == b'':
						await self.in_queue.put(None)
						#print('Connection terminated')
						return
					
					self.ssl_obj.write(raw_data)
					while True:
						try:
							enc_data = self.ssl_out_buff.read()
							if enc_data == b'':
								break
							await self.transport.out_queue.put(enc_data)
						except Exception as e:
							raise e
				except Exception as e:
							raise e

		except asyncio.CancelledError:
			return
		except Exception as e:
			logging.debug('__encrypt_ssl_rec %s' % e)
	
	async def __decrypt_ssl_rec(self):
		"""
		Decrypting the encrypted SSL records from the internal queue and putting them in the queue for the uppr layers to use
		"""
		try:
			while not self.stop_evt.is_set():
				ssl_data = await self.__ssl_in_q.get()
				if ssl_data == b'':
					await self.in_queue.put((ssl_data, None))
					#print('Connection terminated')
					return
				self.ssl_in_buff.write(ssl_data)

				data_buff = b''
				while not self.stop_evt.is_set():
					try:
						data_buff += self.ssl_obj.read()
					except ssl.SSLWantReadError:
						break
					except Exception as e:
						raise e
				#print('data_buff %s' % data_buff)
				if data_buff != b'':
					await self.in_queue.put((data_buff, None))

		except asyncio.CancelledError:
			return
		except Exception as e:
			logging.debug('__decrypt_ssl_rec %s' % e)

	async def __read_ssl_record(self):
		try:
			"""
			buffering the incoming SSL (encrypted) recodrs and putting them in an internal queue one record at a time
			"""
			buffer = b''
			length = None
			while not self.stop_evt.is_set():
				if length is None and len(buffer) >= 6:
					length = int.from_bytes(buffer[3:5], byteorder = 'big', signed = False)
				
				if length is not None and len(buffer) >= length + 5:
					#print('LB raw %s' % len(buffer[:length+5]))
					await self.__ssl_in_q.put(buffer[:length+5])
					buffer = buffer[length+5:]
					length = None
					continue
				
				data, err = await self.transport.in_queue.get()
				if err is not None:
					raise err
				if data == b'':
					await self.__ssl_in_q.put(b'')
					return
				buffer+= data
				

		except asyncio.CancelledError:
			return

		except Exception as e:
			logging.debug('__read_ssl_record %s' % e)
			await self.__ssl_in_q.put(b'')

		#finally:
		#	self.stop_evt.set()
	
	async def __do_handshake(self):
		try:
			ctr = 0
			while not self.stop_evt.is_set():
				ctr += 1
				#print('DST Performing handshake!')
				try:
					self.ssl_obj.do_handshake()
				except ssl.SSLWantReadError:
					#print('DST want %s' % ctr)
					while True:
						client_hello = self.ssl_out_buff.read()
						if client_hello != b'':
							#print('DST client_hello %s' % len(client_hello))
							await self.transport.out_queue.put(client_hello)
						else:
							break
					
					#print('DST wating server hello %s' % ctr)
					server_hello, err = await self.transport.in_queue.get()
					if err is not None:
						raise err
					if server_hello == b'':
						raise Exception('Server closed the connection!')
					#print('DST server_hello %s' % len(server_hello))
					self.ssl_in_buff.write(server_hello)

					continue
				except:
					raise
				else:
					#print('DST handshake ok %s' % ctr)
					#server_fin = tls_out_buff.read()
					#print('DST server_fin %s ' %  server_fin)
					#await out_q.put(server_fin)
					break			

			return True, None

		except Exception as e:
			logging.debug('__do_handshake %s' % e)
			return None, e

	@staticmethod
	async def from_transport(transport, ssl_ctx = None, ssl_cert = None, ssl_key = None):
		try:
			tunnel = SSLClientTunnel()
			tunnel.transport = transport
			if ssl_ctx is not None:
				tunnel.ssl_ctx = ssl_ctx
			elif ssl_key is not None:
				raise NotImplementedError()
			else:
				tunnel.ssl_ctx = ssl.create_default_context()
				tunnel.ssl_ctx.check_hostname = False
				tunnel.ssl_ctx.verify_mode = ssl.CERT_NONE
			_, err = await tunnel.setup()
			if err is not None:
				raise err

			return tunnel, None
		except Exception as e:
			return None, e


from aardwolf.commons.proxy import RDPProxyType
from aardwolf.commons.target import RDPConnectionProtocol, RDPTarget
from aardwolf.transport.tcp import TCPSocket
from aardwolf.network.socks import SocksProxyConnection


class NetworkSelector:
	def __init__(self):
		pass

	@staticmethod
	async def select(target: RDPTarget):
		"""Selects the appropriate network connection library based on the target"""
		try:
			if target.proxy is None:
				if target.protocol == RDPConnectionProtocol.TCP:
					return TCPSocket(target = target), None
				else:
					raise NotImplementedError()
			elif target.proxy.type == RDPProxyType.ASYSOCKS:
				return SocksProxyConnection(target = target), None

			else:
				raise Exception('Cant select correct connection type!')
		except Exception as e:
			return None, e


import traceback
import uuid
import asyncio
import logging
import copy
import json

from aardwolf import logger
from aardwolf.commons.url import RDPConnectionURL
from aardwolf.commons.iosettings import RDPIOSettings
from aardwolf.examples.scancommons.targetgens import *
from aardwolf.examples.scancommons.internal import *
from aardwolf.examples.scancommons.utils import *
from aardwolf.protocol.x224.constants import SUPP_PROTOCOLS
from aardwolf.commons.queuedata.constants import VIDEO_FORMAT
from tqdm import tqdm

class EnumResultFinal:
	def __init__(self, obj, otype, err, target, target_id):
		self.obj = obj
		self.otype = otype
		self.err = err
		self.target = target
		self.target_id = target_id


	def __str__(self):
		if self.err is not None:
			return '[E] %s | %s | %s' % (self.target, self.err)

		elif self.otype == 'result':
			return '[R] %s | %s | %s' % (self.target_id, self.target, self.obj)

		elif self.otype == 'progress':
			return '[P][%s/%s][%s] %s' % (self.obj.total_finished, self.obj.total_targets, str(self.obj.gens_finished), self.obj.current_finished)

		else:
			return '[UNK]'
	
	def to_dict(self):
		if self.otype == 'result':
			return {
				'target' : self.target,
				'target_id' : self.target_id,
				'result': str(self.obj),
				'otype': self.otype,
			}
	
	def to_tsv(self):
		if self.otype == 'result':
			return '%s\t%s\t%s' % (self.target, self.target_id, self.obj)
	
	def to_json(self):
		t = self.to_dict()
		if t is not None:
			return json.dumps(self.to_dict())


class RDPLoginScanner:
	def __init__(self, rdp_url, iosettings:RDPIOSettings, worker_count = 10, out_file = None, out_format = 'str', show_pbar = True, task_q = None, ext_result_q = None):
		self.target_gens = []
		self.rdp_mgr = rdp_url
		if isinstance(rdp_url, RDPConnectionURL) is False:
			self.rdp_mgr = RDPConnectionURL(rdp_url)
		self.worker_count = worker_count
		self.task_q = task_q
		self.res_q = None
		self.workers = []
		self.result_processing_task = None
		self.out_file = out_file
		self.show_pbar = show_pbar
		self.ext_result_q = ext_result_q
		self.enum_url = False
		self.out_format = out_format

		self.iosettings = iosettings
		self.flags_test = [SUPP_PROTOCOLS.HYBRID_EX, None]

		self.__gens_finished = False
		self.__total_targets = 0
		self.__total_finished = 0
		self.__total_errors = 0

	async def __executor(self, tid, target):
		try:
			for i, proto in enumerate(self.flags_test):
				result = 'YES'
				if proto == None:
					result = 'MAYBE'
				ios = copy.deepcopy(self.iosettings)
				ios.supported_protocols = proto
				async with self.rdp_mgr.create_connection_newtarget(target, ios) as connection:
					_, err = await connection.connect()
					if err is not None:
						result = 'NO'
					
					if result == 'NO' and i != (len(self.flags_test)-1):
						# avoid printing NO multiple times
						continue

					await self.res_q.put(EnumResult(tid, target, result, status = EnumResultStatus.RESULT))
					if result == 'YES':
						break

		except asyncio.CancelledError:
			return
		except Exception as e:
			traceback.print_exc()
			await self.res_q.put(EnumResult(tid, target, None, error = e, status = EnumResultStatus.ERROR))
		finally:
			await self.res_q.put(EnumResult(tid, target, None, status = EnumResultStatus.FINISHED))
			

	async def worker(self):
		try:
			while True:
				indata = await self.task_q.get()
				if indata is None:
					return
				
				tid, target = indata
				try:
					await asyncio.wait_for(self.__executor(tid, target), timeout=10)
				except asyncio.CancelledError:
					return
				except asyncio.TimeoutError as e:
					await self.res_q.put(EnumResult(tid, target, None, error = e, status = EnumResultStatus.ERROR))
					await self.res_q.put(EnumResult(tid, target, None, status = EnumResultStatus.FINISHED))
					continue
				except Exception as e:
					logger.exception('worker')
					continue
		except asyncio.CancelledError:
			return
				
		except Exception as e:
			return e

	async def result_processing(self):
		try:
			pbar = None
			if self.show_pbar is True:
				pbar = {}
				pbar['targets']  = tqdm(desc='Targets ', unit='', position=0)
				pbar['yes']      = tqdm(desc='Yes     ', unit='', position=1)
				pbar['no']       = tqdm(desc='No      ', unit='', position=2)
				pbar['maybe']    = tqdm(desc='Maybe   ', unit='', position=3)
				pbar['errors']   = tqdm(desc='Errors  ', unit='', position=4)
				pbar['finished'] = tqdm(desc='Finished', unit='', position=5)

			out_buffer = []
			final_iter = False
			while True:
				try:
					if self.__gens_finished is True and self.show_pbar is True and pbar['targets'].total is None:
						pbar['targets'].total = self.__total_targets
						for key in pbar:
							pbar[key].refresh()

					if self.ext_result_q is not None:
						out_buffer = []

					if len(out_buffer) >= 100 or final_iter and self.ext_result_q is None:
						out_data = ''
						if self.out_format == 'str':
							out_data = '\r\n'.join([str(x) for x in out_buffer])
						elif self.out_format == 'tsv':
							for line in [x.to_tsv() for x in out_buffer]:
								if line is not None:
									out_data += line + '\r\n'
						elif self.out_format == 'json':
							for line in [x.to_json() for x in out_buffer]:
								if line is not None:
									out_data += line + '\r\n'
						if self.out_file is None:
							print(out_data)
						else:
							with open(self.out_file, 'a', newline = '') as f:
								f.write(out_data)
						if self.show_pbar is True:
							for key in pbar:
								pbar[key].refresh()
						
						out_buffer = []
						out_data = ''

					if final_iter:
						asyncio.create_task(self.terminate())
						return
					try:
						er = await asyncio.wait_for(self.res_q.get(), timeout = 5)
					except asyncio.TimeoutError:
						if self.show_pbar is True:
							for key in pbar:
								pbar[key].refresh()

						if self.__total_finished == self.__total_targets and self.__gens_finished is True:
							final_iter = True
						continue
					
					if er.status == EnumResultStatus.FINISHED:
						self.__total_finished += 1
						if self.show_pbar is True:
							pbar['finished'].update(1)

						obj = EnumProgress(self.__total_targets, self.__total_finished, self.__gens_finished, er.target)
						if self.ext_result_q is not None:
							await self.ext_result_q.put(EnumResultFinal(obj, 'progress', None, er.target, er.target_id))
						out_buffer.append(EnumResultFinal(obj, 'progress', None, er.target, er.target_id))
						if self.__total_finished == self.__total_targets and self.__gens_finished is True:
							final_iter = True
							continue
							
					if er.status == EnumResultStatus.RESULT:
						if self.ext_result_q is not None:
							await self.ext_result_q.put(EnumResultFinal(er.result, 'result', None, er.target, er.target_id))
						out_buffer.append(EnumResultFinal(er.result, 'result', None, er.target, er.target_id))
						if self.show_pbar is True:
							if er.result in ['YES', 'NO', 'MAYBE']:
								pbar[er.result.lower()].update(1)
							

					if er.status == EnumResultStatus.ERROR:
						self.__total_errors += 1
						if self.show_pbar is True:
							pbar['errors'].update(1)


				except asyncio.CancelledError:
					return
				except Exception as e:
					logger.exception('result_processing inner')
					asyncio.create_task(self.terminate())
					return
		except asyncio.CancelledError:
			return
		except Exception as e:
			logger.exception('result_processing')
			asyncio.create_task(self.terminate())
		finally:
			if self.ext_result_q is not None:
				await self.ext_result_q.put(EnumResultFinal(None, 'finished', None, None, None))

	async def terminate(self):
		for worker in self.workers:
			worker.cancel()
		if self.result_processing_task is not None:
			self.result_processing_task.cancel()		

	async def setup(self):
		try:
			if self.res_q is None:
				self.res_q = asyncio.Queue(self.worker_count)
				self.result_processing_task = asyncio.create_task(self.result_processing())
			if self.task_q is None:
				self.task_q = asyncio.Queue(self.worker_count)

			for _ in range(self.worker_count):
				self.workers.append(asyncio.create_task(self.worker()))

			return True, None
		except Exception as e:
			return None, e

	async def __generate_targets(self):
		if self.enum_url is True:
			self.__total_targets += 1
			await self.task_q.put((str(uuid.uuid4()), self.rdp_mgr.get_target().get_hostname_or_ip()))
			
		for target_gen in self.target_gens:
			async for uid, target, err in target_gen.generate():
				if err is not None:
					print('Target gen error! %s' % err)
					break
				
				self.__total_targets += 1
				await self.task_q.put((uid, target))
				await asyncio.sleep(0)

		self.__gens_finished = True
	
	async def run(self):
		try:
			_, err = await self.setup()
			if err is not None:
				raise err
			
			gen_task = asyncio.create_task(self.__generate_targets())
			
			await asyncio.gather(*self.workers)
			await self.result_processing_task
			return True, None
		except Exception as e:
			logger.exception('run')
			return None, e

async def amain():
	import argparse
	import sys

	parser = argparse.ArgumentParser(description='RDP login scanner', formatter_class=argparse.RawDescriptionHelpFormatter)
	parser.add_argument('-v', '--verbose', action='count', default=0)
	parser.add_argument('-w', '--worker-count', type=int, default=50, help='Parallell count')
	parser.add_argument('-o', '--out-file', help='Output file path.')
	parser.add_argument('--progress', action='store_true', help='Show progress bar. Pleasue use out-file with this!!')
	parser.add_argument('--tsv', action='store_true', help='Output results in TSV format')
	parser.add_argument('--json', action='store_true', help='Output results in JSON format')
	parser.add_argument('-s', '--stdin', action='store_true', help='Read targets from stdin')
	parser.add_argument('url', help='Connection URL base, target can be set to anything. Example: "rdp+ntlm-password://TEST\\victim:Password!1@10.10.10.2"')
	parser.add_argument('targets', nargs='*', help = 'Hostname or IP address or file with a list of targets')

	args = parser.parse_args()

	if args.verbose >=1:
		logger.setLevel(logging.WARNING)

	if args.verbose > 2:
		print('setting deepdebug')
		logger.setLevel(1) #enabling deep debug
		asyncio.get_event_loop().set_debug(True)
		logging.basicConfig(level=logging.DEBUG)

	oformat = 'str'
	if args.tsv is True:
		oformat = 'tsv'
	if args.json is True:
		oformat = 'json'

	rdp_url = args.url

	iosettings = RDPIOSettings()
	iosettings.channels = []
	iosettings.video_out_format = VIDEO_FORMAT.RAW
	iosettings.clipboard_use_pyperclip = False

	enumerator = RDPLoginScanner(
		rdp_url,
		iosettings,
		worker_count = args.worker_count,
		out_file = args.out_file,
		show_pbar = args.progress,
		out_format=oformat,
	)
	
	notfile = []
	if len(args.targets) == 0 and args.stdin is True:
		enumerator.target_gens.append(ListTargetGen(sys.stdin))
	else:
		for target in args.targets:
			try:
				f = open(target, 'r')
				f.close()
				enumerator.target_gens.append(FileTargetGen(target))
			except:
				notfile.append(target)
	
	if len(notfile) > 0:
		enumerator.target_gens.append(ListTargetGen(notfile))

	if len(enumerator.target_gens) == 0:
		enumerator.enum_url = True

	await enumerator.run()

def main():
	asyncio.run(amain())

if __name__ == '__main__':
	main()


import asyncio
import os

import logging
import typing
from aardwolf.network.tpkt import TPKTNetwork
from aardwolf.protocol.x224.constants import TPDU_TYPE, NEG_FLAGS, SUPP_PROTOCOLS
from aardwolf.protocol.x224.client.connectionrequest import ConnectionRequest
from aardwolf.protocol.x224.server.connectionconfirm import ConnectionConfirm
from aardwolf.protocol.x224.data import Data
from aardwolf.protocol.x224 import X224Packet
from aardwolf.protocol.x224.constants import NEG_FLAGS, SUPP_PROTOCOLS
from aardwolf.protocol.x224.client.connectionrequest import ConnectionRequest, RDP_NEG_REQ
from aardwolf.protocol.x224.server.connectionconfirm import ConnectionConfirm


class X224Network:
	"""
	IMPORTANT! The transpo
	"""
	def __init__(self, transport:TPKTNetwork):
		self.transport:TPKTNetwork = transport

		self.out_queue = asyncio.Queue()
		self.in_queue = asyncio.Queue()
		
		self.disconnected = asyncio.Event()

		self.incoming_task = None
		self.outgoing_task = None
		
	async def disconnect(self):
		if self.outgoing_task is not None:
			self.outgoing_task.cancel()
		if self.incoming_task is not None:
			self.incoming_task.cancel()
		self.disconnected.set()
	
	async def client_negotiate(self, flags:NEG_FLAGS = 0, supported_protocols:SUPP_PROTOCOLS = SUPP_PROTOCOLS.SSL|SUPP_PROTOCOLS.HYBRID_EX, to_raise = True):
		reply = None
		try:
			negreq = RDP_NEG_REQ()
			negreq.flags = flags
			negreq.requestedProtocols = supported_protocols
			
			conreq = ConnectionRequest()
			conreq.SRC_REF = 0
			conreq.cookie = b'Cookie: mstshash=devel\r\n'
			conreq.rdpNegReq = negreq

			#print(X224Packet.from_bytes(conreq.to_bytes()))

			await self.transport.out_queue.put(conreq.to_bytes())
			is_fastpath, reply, err = await self.in_queue.get()
			if err is not None:
				raise err
			if reply.CR != TPDU_TYPE.CONNECTION_CONFIRM:
				raise Exception('Server sent back unknown TPDU type! %s' % reply.CR)
			reply = typing.cast(ConnectionConfirm, reply)
			if reply.rdpNegData is None:
				return reply, None
			
			if reply.rdpNegData.type == 3 and to_raise is True:
				#server denied our request!
				raise Exception('Server refused our connection request! Reason: %s' % reply.rdpNegData.failureCode.name)

			return reply, None
		except Exception as e:
			return reply, e

	async def handle_incoming(self):
		"""
		Reads data bytes from the socket and dispatches it to the incoming queue
		"""
		try:
			lasterror = None
			msgsize = None
			buffer = b''
			while not self.disconnected.is_set():	
				try:
					is_fastpath, data, err = await self.transport.in_queue.get()
					if err is not None:
						raise err
					if data is None:
						return

					packet = data
					if is_fastpath is False:
						#print('X224Net <- %s' % data.hex())
						packet = X224Packet.from_bytes(data)
						#print('X224Net (packet) <- %s' % packet)
					await self.in_queue.put((is_fastpath, packet, None))
				except asyncio.CancelledError as e:
					lasterror = e
					break
				except Exception as e:
					logging.exception('[X224Net] handle_incoming %s' % str(e))
					lasterror = e
					break
			
			
		except asyncio.CancelledError:
			return
		
		except Exception as e:
			lasterror = e

		finally:
			if self.in_queue is not None:
				await self.in_queue.put( (None, None, lasterror) )
			await self.disconnect()
		
	async def handle_outgoing(self):
		"""
		Reads data bytes from the outgoing queue and dispatches it to the socket
		"""
		try:
			while not self.disconnected.is_set():
				data = await self.out_queue.get()
				if data is None:
					return
				msg = Data()
				msg.data = data
				#print('X224Net -> %s' % msg.to_bytes().hex())
				await self.transport.out_queue.put(msg.to_bytes())
		except asyncio.CancelledError:
			return
		except Exception as e:
			logging.debug('[X224Net] handle_outgoing %s' % str(e))
		finally:
			await self.disconnect()
	
	async def run(self):
		try:
			self.incoming_task = asyncio.create_task(self.handle_incoming())
			self.outgoing_task = asyncio.create_task(self.handle_outgoing())
			return True, None
		except Exception as e:
			return False, e

			
			


from ast import Import
import io
import rle
from PIL import Image

from aardwolf.commons.queuedata.constants import VIDEO_FORMAT

bpp_2_bytes = {
	15: 2,
	16: 2,
	24: 3,
	32: 4,
}

def rectconvert(width, height, bitsPerPixel, isCompress, data):
	if bitsPerPixel not in bpp_2_bytes:
		raise ValueError("bitsPerPixel value of %s is not supported!" % bitsPerPixel)
	image = bytes(width * height * 4)
	rle.bitmap_decompress(image, width, height, data, bitsPerPixel, bpp_2_bytes[bitsPerPixel], int(isCompress))
	return Image.frombytes('RGBA', [width, height], image)


import ipaddress
import enum
from urllib.parse import urlparse, parse_qs

from asysocks.common.clienturl import SocksClientURL 


def stru(x):
	return str(x).upper()

class RDPProxySecretType(enum.Enum):
	NONE = 'NONE'
	PLAIN = 'PLAIN'

class RDPProxyType(enum.Enum):
	ASYSOCKS = 'ASYSOCKS'

class RDPProxy:
	def __init__(self):
		self.type:RDPProxyType = None
		self.target = None

	@staticmethod
	def from_params(url_str):
		proxy = RDPProxy()
		url = urlparse(url_str)
		if url.query is None:
			return None

		query = parse_qs(url.query)
		if 'proxytype' not in query and 'sametype' not in query:
			return None

		if query['proxytype'][0].upper() in ['SOCKS5', 'SOCKS4', 'SOCKS4A', 'SOCKS5-SSL', 'SOCKS4-SSL', 'SOCKS4A-SSL', 'WSNET', 'WSNETWS', 'WSNETWSS']:
			proxy.type = RDPProxyType.ASYSOCKS
		else:
			raise NotImplementedError(query['proxytype'][0].upper())
		
		if proxy.type == RDPProxyType.ASYSOCKS:
			cu = SocksClientURL.from_params(url_str)
			cu[-1].endpoint_port = 3389
			proxy.target = cu
		else:
			raise NotImplementedError('Unknown proxy type "%s"' % proxy.type)
		
		return proxy

	def __str__(self):
		t = '==== RDPProxy ====\r\n'
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
			
		return t

def test():
	t = ['socks5://10.10.10.1',
			'socks5+ssl://10.10.10.1',
			'socks5+ssl://admin:password@10.10.10.1',
	]
	for x in t:
		s = RDPProxy.from_connection_string(x)
		print(str(s))

	
if __name__ == '__main__':
	test()


import asyncio

import logging
import traceback
from aardwolf.protocol.tpkt import TPKT

class TPKTNetwork:
	def __init__(self, transport):
		self.transport = transport

		self.out_queue = asyncio.Queue()
		self.in_queue = asyncio.Queue()
		
		self.disconnected = asyncio.Event()

		self.incoming_task = None
		self.outgoing_task = None

		self.__buffer = b''
		self.__soft_switch = False
		
	async def disconnect(self):
		if self.outgoing_task is not None:
			self.outgoing_task.cancel()
		if self.incoming_task is not None:
			self.incoming_task.cancel()
		if self.transport is not None:
			self.transport.out_queue.put_nowait(b'')
			await self.transport.disconnect()
		self.disconnected.set()

	async def handle_incoming(self):
		"""
		Reads data bytes from the socket and dispatches it to the incoming queue
		"""
		try:
			lasterror = None
			msgsize = None
			is_fastpath = False
			while not self.disconnected.is_set():	
				try:
					if msgsize is None:
						if len(self.__buffer) >= 4:
							if self.__buffer[0] == 3:
								msgsize = int.from_bytes(self.__buffer[2:4], byteorder='big', signed=False)
							else:
								is_fastpath = True
								msgsize = self.__buffer[1]
								if msgsize >> 7 == 1:
									msgsize = int.from_bytes(self.__buffer[1:3], byteorder='big', signed=False)
									msgsize = msgsize & 0x7fff


					if msgsize is not None:
						if len(self.__buffer)>=msgsize:
							if is_fastpath is False:
								msg = TPKT.from_bytes(self.__buffer[:msgsize])
								await self.in_queue.put( (is_fastpath, msg.tpdu, None) )
								self.__buffer = self.__buffer[msgsize:]
								msgsize = None
							else:
								await self.in_queue.put( (is_fastpath, self.__buffer[:msgsize], None) )
								self.__buffer = self.__buffer[msgsize:]
								msgsize = None
								is_fastpath = False
						
							if len(self.__buffer) > 0:
								continue

					data, err = await self.transport.in_queue.get()
					if err is not None:
						raise err
					if data is None:
						return

					self.__buffer += data				
				except asyncio.CancelledError as e:
					lasterror = e
					break
				except Exception as e:
					logging.exception('[TPKTNetwork] handle_incoming %s' % str(e))
					lasterror = e
					break
			
			
		except asyncio.CancelledError:
			return
		
		except Exception as e:
			lasterror = e

		finally:
			if self.__soft_switch is False:
				if self.in_queue is not None:
					await self.in_queue.put( (None, None, lasterror) )
				await self.disconnect()
		
	async def handle_outgoing(self):
		"""
		Reads data bytes from the outgoing queue and dispatches it to the socket
		"""
		try:
			while not self.disconnected.is_set():
				data = await self.out_queue.get()
				if data is None:
					return
				msg = TPKT()
				msg.tpdu = data
				#print('TPKT -> %s' % msg.to_bytes())
				await self.transport.out_queue.put(msg.to_bytes())
		except asyncio.CancelledError:
			return
		except Exception as e:
			logging.exception('[TPKTNetwork] handle_outgoing %s' % str(e))
		finally:
			if self.__soft_switch is False:
				await self.disconnect()
	
	async def suspend_read(self):
		try:
			self.__soft_switch = True
			self.incoming_task.cancel()
			self.outgoing_task.cancel()
			return True, None
		except Exception as e:
			return None, e

	async def conitnue_read(self):
		try:
			self.__soft_switch = False
			self.incoming_task = asyncio.create_task(self.handle_incoming())
			self.outgoing_task = asyncio.create_task(self.handle_outgoing())
			return True, None
		except Exception as e:
			return None, e


	async def switch_transport(self, newtransportobj):
		#most cases this means that we are switching to SSL from TCP
		try:
			self.__soft_switch = True #indicating that no errors will be sent to lower/upper layers
			self.incoming_task.cancel()
			self.outgoing_task.cancel()
			self.transport, err = await newtransportobj.from_transport(self.transport)
			if err is not None:
				raise err
			self.__soft_switch = False
			self.incoming_task = asyncio.create_task(self.handle_incoming())
			self.outgoing_task = asyncio.create_task(self.handle_outgoing())
			
			return True, None
		except Exception as e:
			#traceback.print_exc()
			return None, e

	async def run(self):
		try:
			self.incoming_task = asyncio.create_task(self.handle_incoming())
			self.outgoing_task = asyncio.create_task(self.handle_outgoing())
			return True, None
		except Exception as e:
			return False, e

			
			


import enum
from urllib.parse import urlparse, parse_qs
from aardwolf.commons.credential import RDPCredential, RDPCredentialsSecretType, RDPAuthProtocol
from aardwolf.commons.proxy import RDPProxy
from aardwolf.commons.target import RDPTarget, RDPConnectionDialect, RDPConnectionProtocol
from aardwolf.commons.authbuilder import AuthenticatorBuilder
from aardwolf.connection import RDPConnection
from aardwolf.vncconnection import VNCConnection
from getpass import getpass
import base64
import ipaddress
import copy


class RDPConnectionURL:
	def __init__(self, connection_url, target = None, cred = None, proxy = None):
		self.connection_url = connection_url
		
		#credential
		self.authentication_protocol = None
		self.secret_type = None
		self.domain = None
		self.username = None
		self.secret = None
		self.is_anonymous = None
		self.auth_settings = {}
		self.cred = cred

		#target
		self.dialect = None
		self.protocol = None
		self.hostname = None
		self.dc_ip = None
		self.port = None
		self.ip = None
		self.timeout = 5
		self.serverip = None
		self.fragment = None
		self.path = None
		self.compression = False
		self.scheme = None

		self.target = target

		#proxy
		self.proxy = proxy

		if self.connection_url is not None:
			self.parse()

	def get_connection(self, iosettings):
		tio = copy.deepcopy(iosettings)
		credential = self.get_credential()
		target = self.get_target()
		if target.dialect is not None:
			if target.dialect == RDPConnectionDialect.RDP:
				return RDPConnection(target, credential, tio)
			elif target.dialect == RDPConnectionDialect.VNC:
				return VNCConnection(target, credential, tio)
			else:
				raise Exception('Unknown dialect %s' % target.dialect)
		raise Exception('Either target or dialect must be defined first!')

	def create_connection_newtarget(self, ip_or_hostname, iosettings):
		tio = copy.deepcopy(iosettings)
		credential = self.get_credential()
		credential.target = ip_or_hostname

		target = self.get_target()
		try:
			ipaddress.ip_address(ip_or_hostname)
			target.ip = ip_or_hostname
			target.hostname = None
		except:
			target.hostname = ip_or_hostname
			target.ip = ip_or_hostname

		#spneg = AuthenticatorBuilder.to_credssp(credential, target)
		if target.dialect == RDPConnectionDialect.RDP:
			return RDPConnection(target, credential, tio)
		elif target.dialect == RDPConnectionDialect.VNC:
			return VNCConnection(target, credential, tio)
		else:
			raise Exception('Unknown dialect %s' % target.dialect)

	def get_file(self):
		return RDPFile.from_RDPurl(self)

	def get_proxy(self):
		if self.proxy is not None:
			copy.deepcopy(self.proxy)
		return None

	def get_target(self):
		if self.target is not None:
			return copy.deepcopy(self.target)

		if self.ip is not None and self.hostname is None:
			try:
				ipaddress.ip_address(self.ip)
			except:
				self.hostname = self.ip
		if self.serverip is not None:
			self.ip = self.serverip
			
		t = RDPTarget(
			ip = self.ip, 
			port = self.port, 
			hostname = self.hostname, 
			timeout = self.timeout, 
			dc_ip= self.dc_ip, 
			domain = self.domain, 
			proxy = self.get_proxy(),
			protocol=self.protocol,
			dialect=self.dialect
		)
		return t

	def get_credential(self):
		if self.cred is not None:
			return copy.deepcopy(self.cred)
		return RDPCredential(
			username = self.username,
			domain = self.domain, 
			secret = self.secret, 
			secret_type = self.secret_type, 
			authentication_type = self.authentication_protocol, 
			settings = self.auth_settings,
			target = self.ip
		)
	

	def scheme_decoder(self, scheme):
		self.schemes = scheme.upper().split('+')
		
		connection_tags = self.schemes[0].split('-')
		if len(connection_tags) > 1:
			self.dialect = RDPConnectionDialect(connection_tags[0])
			self.protocol = RDPConnectionProtocol(connection_tags[1])
		else:
			self.dialect = RDPConnectionDialect(connection_tags[0])
			self.protocol = RDPConnectionProtocol.TCP

		if len(self.schemes) == 1:
			# user did not specify a requested auth scheme, defaulting to plain
			# later on when server requires credssp this will be converted to NTLM
			self.authentication_protocol = RDPAuthProtocol.PLAIN
			self.secret_type = RDPCredentialsSecretType.PASSWORD
			return

		auth_tags = self.schemes[1].replace('-','_')
		try:
			self.authentication_protocol = RDPAuthProtocol(auth_tags)
		except:
			auth_tags = self.schemes[1].split('-')
			#print(auth_tags)
			if len(auth_tags) > 1:
				if auth_tags[0] == 'MULTIPLEXOR':
					if auth_tags[1] == 'SSL':
						if len(auth_tags) == 2:
							self.authentication_protocol = RDPAuthProtocol.MULTIPLEXOR_SSL_NTLM
						else:
							if auth_tags[2] == 'NTLM':
								self.authentication_protocol = RDPAuthProtocol.MULTIPLEXOR_SSL_NTLM
							elif auth_tags[2] == 'KERBEROS':
								self.authentication_protocol = RDPAuthProtocol.MULTIPLEXOR_SSL_KERBEROS
					else:
						if auth_tags[1] == 'NTLM':
							self.authentication_protocol = RDPAuthProtocol.MULTIPLEXOR_NTLM
						elif auth_tags[1] == 'KERBEROS':
							self.authentication_protocol = RDPAuthProtocol.MULTIPLEXOR_KERBEROS
				elif auth_tags[0] == 'SSPI':
					if auth_tags[1] == 'NTLM':
						self.authentication_protocol = RDPAuthProtocol.SSPI_NTLM
					elif auth_tags[1] == 'KERBEROS':
						self.authentication_protocol = RDPAuthProtocol.SSPI_KERBEROS
				else:
					self.authentication_protocol = RDPAuthProtocol(auth_tags[0])
					self.secret_type = RDPCredentialsSecretType(auth_tags[1])
			else:
				if auth_tags[0] == 'MULTIPLEXOR':
					self.authentication_protocol = RDPAuthProtocol.MULTIPLEXOR_NTLM
				elif auth_tags[0] == 'MULTIPLEXOR_SSL':
					self.authentication_protocol = RDPAuthProtocol.MULTIPLEXOR_SSL_NTLM
				if auth_tags[0] == 'SSPI':
					self.authentication_protocol = RDPAuthProtocol.SSPI_NTLM
				else:
					self.authentication_protocol = RDPAuthProtocol(auth_tags[0])
				if self.authentication_protocol == RDPAuthProtocol.KERBEROS:
					raise Exception('For kerberos auth you need to specify the secret type in the connection string!')

	def __str__(self):
		t = '==== RDPConnectionURL ====\r\n'
		for k in self.__dict__:
			val = self.__dict__[k]
			if isinstance(val, enum.IntFlag):
				val = val
			elif isinstance(val, enum.Enum):
				val = val.name
			
			t += '%s: %s\r\n' % (k, str(val))
			
		return t

	def parse(self):
		url_e = urlparse(self.connection_url)
		self.scheme_decoder(url_e.scheme)
		
		if url_e.username is not None:
			if url_e.username.find('\\') != -1:
				self.domain , self.username = url_e.username.split('\\')
				if self.domain == '.':
					self.domain = None
			else:
				self.domain = None
				self.username = url_e.username
		
		self.secret = url_e.password
		
		if self.secret_type == RDPCredentialsSecretType.PWPROMPT:
			self.secret_type = RDPCredentialsSecretType.PASSWORD
			self.secret = getpass()

		if self.secret_type == RDPCredentialsSecretType.PWHEX:
			self.secret_type = RDPCredentialsSecretType.PASSWORD
			self.secret = bytes.fromhex(self.secret).decode()
		
		if self.secret_type == RDPCredentialsSecretType.PWB64:
			self.secret_type = RDPCredentialsSecretType.PASSWORD
			self.secret = base64.b64decode(self.secret).decode()
		
		if self.secret is None and self.username is None:
			self.is_anonymous = True
		
		if self.authentication_protocol == RDPAuthProtocol.NTLM and self.secret_type is None:
			if self.is_anonymous == True:
				self.secret_type = RDPCredentialsSecretType.NONE
			else:
				if len(self.secret) == 32:
					try:
						bytes.fromhex(self.secret)
					except:
						self.secret_type = RDPCredentialsSecretType.NT
					else:
						self.secret_type = RDPCredentialsSecretType.PASSWORD

		elif self.authentication_protocol in [RDPAuthProtocol.SSPI_KERBEROS, RDPAuthProtocol.SSPI_NTLM, 
												RDPAuthProtocol.MULTIPLEXOR_NTLM, RDPAuthProtocol.MULTIPLEXOR_KERBEROS]:
			if self.username is None:
				self.username = '<CURRENT>'
			if self.domain is None:
				self.domain = '<CURRENT>'
			if self.secret is None:
				self.secret = '<CURRENT>'


		self.ip = url_e.hostname
		if url_e.port:
			self.port = url_e.port
		elif self.protocol == RDPConnectionProtocol.TCP and self.dialect == RDPConnectionDialect.RDP:
			self.port = 3389
		elif self.protocol == RDPConnectionProtocol.TCP and self.dialect == RDPConnectionDialect.VNC:
			self.port = 5800
		#elif self.protocol == RDPConnectionProtocol.QUIC:
		#	self.port = 443
		else:
			raise Exception('Port must be provided!')

		if url_e.path not in ['/', '', None]:
			self.path = url_e.path
		

		# recognized parameters :
		# dc -> domain controller IP
		# proxytype -> proxy protocol
		# proxyuser -> username for proxy auth
		# proxypass -> password for proxy auth
		#  
		#
		proxy_present = False
		if url_e.query is not None:
			query = parse_qs(url_e.query)
			for k in query:
				if k.startswith('proxy') is True:
					proxy_present = True
				if k == 'dc':
					self.dc_ip = query[k][0]
				elif k == 'timeout':
					self.timeout = int(query[k][0])
				elif k == 'serverip':
					self.serverip = query[k][0]
				elif k == 'fragment':
					self.fragment = int(query[k][0])
				elif k == 'dns':
					self.dns = query[k] #multiple dns can be set, so not trimming here
				elif k == 'compress':
					self.compression = int(query[k][0])
				elif k.startswith('auth'):
					self.auth_settings[k[len('auth'):]] = query[k]
				elif k.startswith('same'):
					self.auth_settings[k[len('same'):]] = query[k]
		
		if proxy_present is True:
			self.proxy = RDPProxy.from_params(self.connection_url)
			
if __name__ == '__main__':
	from aardwolf.commons.interfaces.file import RDPFile
	url_tests = [
		#'RDP://10.10.10.2',
		#'RDP://10.10.10.2:9000',
		#'RDP-tcp://10.10.10.2',
		#'RDP-tcp://10.10.10.2:9000',
		#'RDP-udp://10.10.10.2:138',
		#'RDP+ntlm-password://domain\\user@10.10.10.2',
		#'RDP-tcp+ntlm-password://domain\\user:password@10.10.10.2',
		#'RDP-tcp+ntlm-password://domain\\user:password@10.10.10.2:10000',
		#'RDP-tcp+ntlm-nt://domain\\user:alma@10.10.10.2',
		#'RDP+ntlm-nt://domain\\user:alma@10.10.10.2',
		#'RDP+ntlm-nt://domain\\user:alma@10.10.10.2',
		#'RDP-tcp+kerberos-password://domain\\alma:password@10.10.10.10.2',
		#'RDP-tcp+kerberos-aes://domain\\alma:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa@10.10.10.10.2',
		#'RDP-tcp+kerberos-aes://domain\\alma:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa@10.10.10.10.2',
		#'RDP-tcp+kerberos-nt://domain\\alma:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa@10.10.10.10.2',
		#'RDP-tcp+kerberos-rc4://domain\\alma:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa@10.10.10.10.2',
		#'RDP+sspi://10.10.10.10.2',
		#'RDP+sspi-ntlm://10.10.10.10.2',
		#'RDP+sspi-kerberos://10.10.10.10.2',
		#'RDP+multiplexor://10.10.10.10.2',
		#'RDP+multiplexor-ssl://10.10.10.10.2',
		#'RDP+multiplexor-ssl-ntlm://10.10.10.10.2',
		#'RDP+multiplexor-ssl-kerberos://10.10.10.10.2',
		#'RDP://10.10.10.2/?timeout=10',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=socks5',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=socks5&proxyserver=127.0.0.1',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=socks5&proxyserver=127.0.0.1&proxyuser=admin&proxypass=alma',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=socks5&proxyserver=127.0.0.1&proxyuser=admin&proxypass=alma&dc=10.10.10.2&dns=8.8.8.8',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=socks5-ssl&proxyserver=127.0.0.1&proxyuser=admin&proxypass=alma&dc=10.10.10.2&dns=8.8.8.8',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=multiplexor&proxyserver=127.0.0.1',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=multiplexor&proxyserver=127.0.0.1&proxyagentid=alma',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=multiplexor&proxyserver=127.0.0.1&proxyagentid=alma&proxytimeout=111',
		'RDP://10.10.10.2/C$/test/tst111.dmp?timeout=10&dc=10.10.10.2&proxytype=multiplexor&proxyhost=127.0.0.1&proxyport=1&proxyagentid=alma&proxytimeout=111',

	]
	for url in url_tests:
		print('===========================================================================')
		print(url)
		try:
			dec = RDPConnectionURL(url)
			creds = dec.get_credential()
			target = dec.get_target()
			RDPfile = dec.get_file()
			print(RDPfile)
		except Exception as e:
			import traceback
			traceback.print_exc()
			print('ERROR! Reason: %s' % e)
			input()
		else:
			print(str(creds))
			print(str(target))
			input()
			
			




import asyncio
import traceback
import typing

from aardwolf.protocol.T124.userdata.constants import ChannelOption

class Channel:
	def __init__(self, name:str, options:ChannelOption = ChannelOption.INITIALIZED|ChannelOption.ENCRYPT_RDP):
		self.name = name
		self.options = options
		self.channel_id = None
		self.raw_in_queue = asyncio.Queue()
		self.raw_out_queue = None
		self.in_queue = asyncio.Queue()
		self.data_in_queue = asyncio.Queue()
		self.out_queue = asyncio.Queue()
		self.connection = None
		self.initiator = None

	def get_name(self):
		return self.name
	
	async def disconnect(self):
		try:
			if self.monitor_out_task is not None:
				self.monitor_out_task.cancel()
			if self.monitor_in_task is not None:
				self.monitor_in_task.cancel()
			_, err = await self.stop()
			if err is not None:
				raise err
		except:
			pass

	async def run(self, connection):
		try:
			# this should take a queues which belong to the lower-layer
			self.raw_out_queue = connection._x224net.out_queue
			self.connection = connection
			self.initiator = self.connection._initiator
			self.monitor_out_task = asyncio.create_task(self.monitor_out())
			self.monitor_in_task = asyncio.create_task(self.monitor_in())
			_, err = await self.start()
			if err is not None:
				raise err
			return True, None
		except Exception as e:
			traceback.print_exc()
			return None, e


	async def start(self):
		# Override this when implementing new channel type
		try:
			return True, None
		except Exception as e:
			return None, e
	
	async def stop(self):
		# Override this when implementing new channel type
		try:
			return True, None
		except Exception as e:
			return None, e

	async def monitor_in(self):
		try:
			while True:
				data, err = await self.raw_in_queue.get()						
				if err is not None:
					await self.out_queue.put((data, err))
					raise err
				#print('Channel data in! "%s(%s)" <- %s' % (self.name, self.channel_id, data))

				await self.out_queue.put((data, err))

		except asyncio.CancelledError:
			return None, None
		except Exception as e:
			traceback.print_exc()
			return None, e
	
	async def monitor_out(self):
		try:
			while True:
				data = await self.in_queue.get()
				#print('Sending data on channel "%s(%s)" -> %s' % (self.name, self.channel_id, data))
				await self.raw_out_queue.put(data)

		except asyncio.CancelledError:
			return None, None

		except Exception as e:
			traceback.print_exc()
			return None, e
	
	



import io
import enum

class CLIPRDR_TEMP_DIRECTORY:
	def __init__(self):
		self.wszTempDir:str = None #always 520 bytes

	def to_bytes(self):
		t = self.wszTempDir.encode('utf-16-le').ljust(520,b'\x00')
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return CLIPRDR_TEMP_DIRECTORY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = CLIPRDR_TEMP_DIRECTORY()
		msg.wszTempDir = buff.read(520).decode('utf-16-le').replace('\x00','')
		return msg

	def __repr__(self):
		t = '==== CLIPRDR_TEMP_DIRECTORY ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import enum
import asyncio
import ipaddress

from aardwolf import logger

from asysocks.client import SOCKSClient
from asysocks.common.comms import SocksQueueComms


class SocksProxyConnection:
	"""
	Generic asynchronous TCP socket class, nothing SMB related.
	Creates the connection and channels incoming/outgoing bytes via asynchonous queues.
	"""
	def __init__(self, target):
		self.target = target
		
		self.client = None
		self.proxy_task = None
		self.handle_in_task = None

		self.out_queue = None#asyncio.Queue()
		self.in_queue = None#asyncio.Queue()
		self.disconnected = None #asyncio.Event()

		self.proxy_in_queue = None#asyncio.Queue()
		self.is_plain_msg = True
		
	async def disconnect(self):
		"""
		Disconnects from the socket.
		Stops the reader and writer streams.
		"""
		if self.client is not None:
			await self.client.terminate()
		if self.proxy_task is not None:
			self.proxy_task.cancel()
		if self.handle_in_q is not None:
			self.handle_in_task.cancel()

	async def terminate(self):
		await self.disconnect()
			
	async def handle_in_q(self):
		try:
			while True:				
				temp, err = await self.proxy_in_queue.get()
				#print(temp)
				if err is not None:
					raise err

				if temp == b'' or temp is None:
					logger.debug('Server finished!')
					return

				await self.in_queue.put((temp, None))
				continue
		
		except asyncio.CancelledError:
			return
		except Exception as e:
			logger.exception('handle_in_q')
			await self.in_queue.put((None, e))

		finally:
			self.proxy_task.cancel()


		
	async def connect(self):
		"""
		
		"""
		try:
			self.out_queue = asyncio.Queue()
			self.in_queue = asyncio.Queue()
			self.disconnected = asyncio.Event()
			self.proxy_in_queue = asyncio.Queue()
			comms = SocksQueueComms(self.out_queue, self.proxy_in_queue)

			self.target.proxy.target[-1].endpoint_ip = self.target.get_hostname_or_ip() if self.target.serverip is None else self.target.serverip
			self.target.proxy.target[-1].endpoint_port = int(self.target.port)
			self.target.proxy.target[-1].endpoint_timeout = None #TODO: maybe implement endpoint timeout?
			self.target.proxy.target[-1].timeout = self.target.timeout
			self.client = SOCKSClient(comms, self.target.proxy.target)
			self.proxy_task = asyncio.create_task(self.client.run())
			self.handle_in_task = asyncio.create_task(self.handle_in_q())
			return True, None
		except Exception as e:
			return False, e

